class FLOJOY_INSTRUCTION:
    FLOW_TO_DIRECTIONS = "__flow_to_directions__"
    FLOW_TO_NODES = "__flow_to_nodes__"
    RESULT_FIELD = "__result_field__"
    DATACONTAINER_FIELD = "__data_container_field__"
